import gazu

from kabaret import flow

from .maputils import ClearMapAction
from kabaret.subprocess_manager.flow import RunAction


class KitsuAPIWrapper(flow.Object):

    _server_url = flow.Param("")
    _config = flow.Parent()

    def set_host(self, url):
        gazu.client.set_host(url)

    def get_host(self):
        return gazu.client.get_host()

    def set_server_url(self, url):
        self._server_url.set(url)

    def get_server_url(self):
        return self._server_url.get()

    def log_in(self, username, password):
        try:
            gazu.log_in(username, password)
        except (
            gazu.exception.AuthFailedException,
            gazu.exception.ServerErrorException,
        ):
            return False

        return True
    
    def log_out(self):
        gazu.log_out()

    def get_tokens(self):
        return gazu.client.default_client.tokens

    def set_tokens(self, tokens):
        gazu.client.set_tokens(tokens)

    def host_is_valid(self):
        if not gazu.client.host_is_up(gazu.client.default_client):
            return False
        try:
            gazu.client.post("auth/login", {"email": "", "password": ""})
        except Exception as exc:
            return (
                type(exc) == gazu.exception.ParameterException
                or type(exc) == gazu.exception.ServerErrorException
            )

    def current_user_logged_in(self):
        """
        Checks if the current user is logged in.

        This method assumes Kitsu client's host is valid.
        """
        try:
            gazu.client.get_current_user()
        except gazu.exception.NotAuthenticatedException:
            return False

        return True
    
    def get_project_data(self):
        return gazu.project.get_project_by_name(
            self._config.project_name.get()
        )
    
    def get_shot_data(self, name, sequence):
        if isinstance(sequence, str):
            sequence = self.get_sequence_data(sequence)

        if not sequence:
            return None
        
        return gazu.shot.get_shot_by_name(
            sequence,
            name
        )
    
    def get_shot_casting(self, shot, sequence=None):
        if isinstance(shot, str):
            shot = self.get_shot_data(shot, sequence)

        if not shot:
            return None

        return gazu.casting.get_shot_casting(shot)
    
    def get_sequence_data(self, name):
        return gazu.shot.get_sequence_by_name(
            self._config.project_id.get(),
            name
        )
    
    def get_sequence_casting(self, sequence):
        if isinstance(sequence, str):
            sequence = self.get_sequence_data(sequence)

        if not sequence:
            return None

        return gazu.casting.get_sequence_casting(sequence)
    
    def get_asset_data(self, name):
        return gazu.asset.get_asset_by_name(
            self._config.project_id.get(),
            name
        )
    
    def get_asset_type(self, asset):
        if isinstance(asset, str):
            asset = self.get_asset_data(asset)

        if not asset:
            return None

        return gazu.asset.get_asset_type(asset["entity_type_id"])


class KitsuBindings(flow.Object):

    asset_types = flow.HashParam()
    asset_families = flow.HashParam()

    def _kitsu_api(self):
        return self.root().project().kitsu_api()

    def get_asset_oid(self, kitsu_asset_name):
        kitsu_api = self.root().project().kitsu_api()
        kitsu_asset = kitsu_api.get_asset_data(kitsu_asset_name)

        asset_type = kitsu_api.get_asset_type(kitsu_asset['name'])['name']
        asset_type = self.get_asset_type(asset_type)
        asset_family = self.get_asset_family(kitsu_asset['data']['family'])

        return '%s/asset_lib/asset_types/%s/asset_families/%s/assets/%s' % (
            self.root().project().oid(),
            asset_type,
            asset_family,
            kitsu_asset['name']
        )

    def get_asset_data(self, name):
        kitsu_api = self._kitsu_api()
        kitsu_asset = kitsu_api.get_asset_data(name)

        return dict(
            type=kitsu_api.get_asset_type(name)['name'],
            family=kitsu_asset['data']['family']
        )

    def get_shot_casting(self, shot_name, sequence_name):
        kitsu_api = self._kitsu_api()
        kitsu_casting = kitsu_api.get_shot_casting(shot_name, sequence_name)
        casting = dict()

        for asset in kitsu_casting:
            asset_name = asset['asset_name']
            asset_data = self.get_asset_data(asset_name)
            asset_data['nb_occurrences'] = asset['nb_occurences']
            casting[asset_name] = asset_data

        return casting

    def get_asset_type(self, kitsu_asset_type):
        if not self.asset_types.has_key(kitsu_asset_type):
            return kitsu_asset_type

        return self.asset_types.get_key(kitsu_asset_type)

    def get_asset_family(self, kitsu_asset_family):
        if not self.asset_families.has_key(kitsu_asset_family):
            return kitsu_asset_family

        return self.asset_families.get_key(kitsu_asset_family)

class UpdateKitsuSettings(flow.Action):

    _kitsu_object = flow.Parent(2)

    def needs_dialog(self):
        return False

    def run(self, button):
        self._kitsu_object.update_kitsu_settings()


class UpdateItemsKitsuSettings(flow.Action):

    _kitsu_map = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        for item in self._kitsu_map.mapped_items():
            item.update_kitsu_settings()

        self._kitsu_map.touch()


class KitsuSetting(flow.values.Value):
    pass


class KitsuSettings(flow.Map):

    clear_settings = flow.Child(ClearMapAction)
    update_settings = flow.Child(UpdateKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return KitsuSetting

    def columns(self):
        return ["Name", "Value"]

    def _fill_row_cells(self, row, item):
        row["Name"] = item.name()
        row["Value"] = item.get()

    def update(self, settings):
        try:
            settings["kitsu_name"] = settings.pop("name")
        except KeyError:
            pass

        for name, value in settings.items():
            try:
                kitsu_setting = self.get_mapped(name)
            except flow.exceptions.MappedNameError:
                kitsu_setting = self.add(name)

            kitsu_setting.set(value)

        self.touch()


class OpenInBrowser(RunAction):

    ICON = ("icons.libreflow", "firefox")

    _url = flow.Parent()

    def runner_name_and_tags(self):
        return "Firefox", ["Browser"]

    def extra_argv(self):
        return [self._url.get()]

    def allow_context(self, context):
        return context and context.endswith(".inline")

    def needs_dialog(self):
        return False


class Url(flow.values.ComputedValue):

    open_in_browser = flow.Child(OpenInBrowser)


class KitsuObject(flow.Object):
    """
    Abstract class representing a Kitsu entity.

    Subclasses must implement the *kitsu_dict* and *compute_child_value* methods.
    """

    kitsu_settings = flow.Child(KitsuSettings).ui(hidden=True)
    kitsu_url = flow.Computed(computed_value_type=Url).ui(hidden=True)
    kitsu_id = flow.Param().ui(editable=False).ui(hidden=True)

    def kitsu_setting_names(self):
        """
        Returns the list of object's settings names, as a subset of the keys
        of the dictionary returned by *kitsu_dict*.

        Returning None will skip name filtering on *kitsu_dict* result
        in *get_kitsu_settings*.
        """
        return None

    def kitsu_dict(self):
        """
        Must be implemented to return a dictionary of parameters related
        to the Kitsu entity.

        It should simply consists in calling the appropriate Gazu
        function given the object's *kitsu_id*.
        """
        raise NotImplementedError()

    def get_kitsu_settings(self):
        settings = self.kitsu_dict()
        names = self.kitsu_setting_names()

        if names is None:
            return settings

        return {name: settings[name] for name in names}

    def update_kitsu_settings(self):
        self.kitsu_settings.update(self.get_kitsu_settings())


class KitsuMap(flow.Map):
    @classmethod
    def mapped_type(cls):
        return KitsuObject


class EntityType(flow.values.ChoiceValue):

    CHOICES = ["Assets", "Shots"]


class SyncFromKitsu(flow.Action):

    ICON = ("icons.libreflow", "sync_arrow")

    entity_type = flow.Param("Assets", EntityType)
    from_index = flow.IntParam(0).ui(label="From")
    to_index = flow.IntParam(10).ui(label="To")

    def get_buttons(self):
        self.message.set("<h3>Synchronize entities from Kitsu</h3>")

        return ["Synchronize", "Cancel"]

    def run(self, button):
        if button == "Cancel":
            return

        project = self.root().project()
        project_kitsu_id = project.kitsu_id.get()

        import time

        start_time = time.time()
        i = 0

        if self.entity_type.get() == "Shots":
            kitsu_sequences = gazu.shot.all_sequences_for_project(project_kitsu_id)[
                self.from_index.get() : self.to_index.get()
            ]
            sequences = project.sequences

            # Pull sequences
            for kitsu_sequence in kitsu_sequences:
                try:
                    sequence = sequences.add(kitsu_sequence["name"])
                except flow.exceptions.MappedNameError:
                    # Ignore sequence already mapped
                    continue

                sequence_id = kitsu_sequence["id"]
                sequence.kitsu_id.set(sequence_id)
                sequence.description.set(kitsu_sequence["description"])
                sequence.update_kitsu_settings()

                # Pull shots
                kitsu_shots = gazu.shot.all_shots_for_sequence(sequence_id)
                shots = sequence.shots

                for kitsu_shot in kitsu_shots:
                    try:
                        shot = shots.add(kitsu_shot["name"])
                    except flow.exceptions.MappedNameError:
                        # Ignore shot already mapped
                        continue

                    shot.kitsu_id.set(kitsu_shot["id"])
                    shot.description.set(kitsu_shot["description"])
                    shot.update_kitsu_settings()

                    i += 1

                shots.touch()

            sequences.touch()

            elapsed_time = float(time.time() - start_time)
            self.root().session().log_debug(
                "Elapsed time: {:.4f} min. ({:.4f} min. per shot) ({} shots)".format(
                    elapsed_time / 60.0, elapsed_time / (60.0 * float(i)), i
                )
            )

        elif self.entity_type.get() == "Assets":
            kitsu_assets = gazu.asset.all_assets_for_project(project_kitsu_id)[
                self.from_index.get() : self.to_index.get()
            ]
            assets = project.asset_lib
            i = 0

            for kitsu_asset in kitsu_assets:
                try:
                    asset = assets.add(kitsu_asset["name"])
                except (flow.exceptions.MappedNameError, TypeError) as e:
                    if isinstance(e, flow.exceptions.MappedNameError):
                        # Asset is already mapped
                        i += 1
                        continue

                    try:
                        asset = assets.add("asset{:04d}".format(i))
                    except flow.exceptions.MappedNameError:
                        i += 1
                        continue

                asset.kitsu_id.set(kitsu_asset["id"])
                asset.description.set(kitsu_asset["description"])
                asset.update_kitsu_settings()

                i += 1

            assets.touch()


class KitsuProject(KitsuObject):

    kitsu_name = flow.Param("").watched().ui(hidden=True)
    kitsu_url = flow.Computed().ui(hidden=True)
    kitsu_id = flow.Computed().ui(hidden=True)

    kitsu_api = flow.Child(KitsuAPIWrapper).ui(hidden=True)
    sync_from_kitsu = flow.Child(SyncFromKitsu).injectable().ui(label="Synchronize", hidden=True)

    def kitsu_dict(self):
        project_name = self.kitsu_name.get()
        if not project_name:
            project_name = self.name()

        return gazu.project.get_project_by_name(project_name)

    def child_value_changed(self, child_value):
        if child_value is self.kitsu_name:
            self.kitsu_id.touch()
            self.kitsu_url.touch()

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_id:
            project_dict = self.kitsu_dict()
            child_value.set(project_dict["id"])
        elif child_value is self.kitsu_url:
            child_value.set(
                "%s/productions/%s"
                % (self.kitsu_api.get_server_url(), self.kitsu_id.get())
            )


class KitsuShot(KitsuObject):

    # def kitsu_setting_names(self):
    #     return ['name', 'description', 'nb_frames', 'data']

    def kitsu_dict(self):
        return gazu.shot.get_shot(self.kitsu_id.get())

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/shots/%s"
                % (self.root().project().kitsu_url.get(), self.kitsu_id.get())
            )


class KitsuSequence(KitsuObject):
    def kitsu_dict(self):
        return gazu.shot.get_sequence(self.kitsu_id.get())

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/shots?search=%s"
                % (self.root().project().kitsu_url.get(), self.name())
            )


class KitsuAsset(KitsuObject):
    def kitsu_dict(self):
        return gazu.asset.get_asset(self.kitsu_id.get())

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/assets/%s"
                % (self.root().project().kitsu_url.get(), self.kitsu_id.get())
            )



import gazu
from kabaret import flow

from libreflow import baseflow


class EntityType(flow.values.ChoiceValue):

    CHOICES = ["Assets", "Episodes"]


class SyncFromKitsu(baseflow.kitsu.SyncFromKitsu):

    ICON = ("icons.libreflow", "sync_arrow")

    entity_type = flow.Param("Assets", EntityType)
    from_index = flow.IntParam().ui(hidden=True)
    to_index = flow.IntParam().ui(hidden=True)

    def run(self, button):
        if button == "Cancel":
            return

        project = self.root().project()
        project_kitsu_id = project.kitsu_id.get()

        if self.entity_type.get() == "Episodes":
            kitsu_episodes = gazu.shot.all_episodes_for_project(project_kitsu_id)
            episodes = project.episodes

            # Pull episodes
            for kitsu_episode in kitsu_episodes:
                try:
                    episode = episodes.add(kitsu_episode["name"])
                except ValueError:
                    episode = episodes[kitsu_episode["name"]]

                episode_id = kitsu_episode["id"]
                episode.kitsu_id.set(episode_id)
                episode.update_kitsu_settings()

                # Pull sequences
                kitsu_sequences = gazu.shot.all_sequences_for_episode(episode_id)
                sequences = episode.sequences

                for kitsu_sequence in kitsu_sequences:
                    try:
                        sequence = sequences.add(kitsu_sequence["name"])
                    except ValueError:
                        sequence = sequences[kitsu_sequence["name"]]

                    sequence_id = kitsu_sequence["id"]
                    sequence.kitsu_id.set(sequence_id)
                    sequence.description.set(kitsu_sequence["description"])
                    sequence.update_kitsu_settings()

                    # Pull shots
                    kitsu_shots = gazu.shot.all_shots_for_sequence(sequence_id)
                    shots = sequence.shots

                    for kitsu_shot in kitsu_shots:
                        try:
                            shot = shots.add(kitsu_shot["name"])
                        except ValueError:
                            shot = shots[kitsu_shot["name"]]

                        shot.kitsu_id.set(kitsu_shot["id"])
                        shot.description.set(kitsu_shot["description"])
                        shot.update_kitsu_settings()

                    shots.touch()

                sequences.touch()

            episodes.touch()

        elif self.entity_type.get() == "Assets":
            kitsu_assets = gazu.asset.all_assets_for_project(project_kitsu_id)
            asset_types = project.asset_lib.asset_types

            for kitsu_asset in kitsu_assets:
                kitsu_asset = gazu.asset.get_asset(kitsu_asset["id"])
                asset_type_name = kitsu_asset["asset_type_name"]
                asset_name = kitsu_asset["name"]

                try:
                    asset_type = asset_types.add(asset_type_name)
                except ValueError:
                    asset_type = asset_types[asset_type_name]

                assets = asset_type.assets

                try:
                    asset = assets.add(asset_name)
                except ValueError:
                    asset = assets[asset_name]

                asset.kitsu_id.set(kitsu_asset["id"])
                asset.description.set(kitsu_asset["description"])
                asset.update_kitsu_settings()

            asset_types.touch()


class KitsuEpisode(baseflow.kitsu.KitsuObject):
    def kitsu_dict(self):
        return gazu.shot.get_episode(self.kitsu_id.get())

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/episodes/%s/shots"
                % (self.root().project().kitsu_url.get(), self.kitsu_id.get())
            )



import os
import shutil
import re

from kabaret import flow

class RemoveGenericAction(flow.Action):
    """
    This action delete the object from its parent's map.
    Then the goto sends the user back to the map has it should no be on the
    object anymore.
    """
    _parentGoto = flow.Parent(3)
    _map = flow.Parent(2)
    _entity = flow.Parent()
    ICON = ('icons.gui', "remove-button")

    def get_buttons(self):
        self.message.set(
            "Are you REALLY sure you want to delete : %s" %
            self._entity.name())
        return ['YES I AM SURE']

    def run(self, button):
        map = self._map
        parentGoto = self._parentGoto
        map.remove(self._entity.name())
        map.touch()
        return self.get_result(goto=parentGoto.oid())


class CreateGenericAction(flow.Action):
    """
    This is a simple create action for maps. It's usefull only for
    simple creations with just the name of the object.
    You should create a custom create action if you need mor information
    """
    _map = flow.Parent()
    entity_name = flow.Param('')

    def get_buttons(self):
        return ['Create']

    def run(self, button):
        if self.entity_name.get() == "":
            print('Action not executed as you have to provide an entity name')
            return
        self._map.add(self.entity_name.get())
        self._map.touch()


class CreateGenericItemAction(flow.Action):
    """
    Creates an item with a generic key.

    This class may be parented to a *GenericMap* (or subclass),
    which provides items prefix and index format padding to
    define their keys in the map.
    """

    ICON = ("icons.gui", "plus-sign-in-a-black-circle")

    _map = flow.Parent()

    def get_buttons(self):
        item_class_name = self._map.mapped_type().__name__.lower()
        self.message.set("<h2>Create %s</h2>" % item_class_name)

        return ["Create", "Cancel"]

    def needs_dialog(self):
        return False

    def run(self, button):
        if button == "Cancel":
            return

        map_size = len(self._map)
        name = "{prefix}{item_index:0{padding}d}".format(
            prefix=self._map.item_prefix,
            item_index=map_size,
            padding=self._map.item_padding,
        )
        item = self._map.add(name)

        self._map.touch()

        return item


class SimpleCreateAction(CreateGenericItemAction):
    """
    This is a simple create action for maps. It's usefull only for
    simple creations with just the name of the object.
    You should create a custom create action if you need more information
    """

    ICON = ("icons.gui", "plus-sign-in-a-black-circle")

    entity_name = flow.Param("").ui(label="Name")

    def needs_dialog(self):
        return True

    def run(self, button):
        if button == "Cancel":
            return

        if self.entity_name.get() == "":
            msg = self.message.get()
            msg += (
                "<font color=#D5000D>%s name must be not empty.</font>"
                % self._map.mapped_type().__name__.lower()
            )
            self.message.set(msg)

            return self.get_result(close=False)

        self._map.add(self.entity_name.get())
        self._map.touch()


class CreateItemAction(CreateGenericItemAction):
    """
    Adds to a map an item containing a description param.
    """

    description = flow.Param("")

    def needs_dialog(self):
        return True

    def run(self, button):
        if button == "Cancel":
            return

        super(CreateItemAction, self).run(button)
        item = self._map.mapped_items()[-1]
        item.description.set(self.description.get())

        self._map.touch()


class ClearMapAction(flow.Action):
    """
    Clears all map's items.
    """

    ICON = ("icons.gui", "remove-symbol")

    _map = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        self._map.clear()
        self._map.touch()


class GenericItemMap(flow.Map):
    """
    A map defining attributes used to index items generically.

    An item's key is made of a prefix (*item_prefix*) and the item's
    index formatted according to a padding number (*item_padding*).
    Subclasses may provide a relation to a *CreateGenericItemAction*
    (or subclass) making use of these attributes.
    """

    item_prefix = "prefix_"
    item_padding = 3


class ItemMap(GenericItemMap):
    """
    A map composed of items with a description.
    """

    def columns(self):
        return ["Name", "Description"]

    def _fill_row_cells(self, row, item):
        row["Name"] = item.name()
        row["Description"] = item.description.get()



import os
import sys
from minio import Minio


server_url = sys.argv[1]
access_key = sys.argv[2]
secret_key = sys.argv[3]
local_path = sys.argv[4]
server_path = sys.argv[5]

minioClient = Minio(server_url,
                  access_key=access_key,
                  secret_key=secret_key,
                  secure=True)

minioClient.fput_object(
    "testbucked",
    server_path,
    local_path
)
